#ifndef _ASSIGN_INV_MATVEC_MIC_H
#define _ASSIGN_INV_MATVEC_MIC_H

#ifdef _CLOVER
template <typename T_solver, typename T_wqf, typename T_gvf, typename T_clvf>
void
assign_inv_matvec(       T_solver  &solver,
                   const float     *skappa,
                   const float     *stol,
                         int       *iter,
                         T_wqf     *restrict yd,
                   const T_wqf     *restrict  y,
                   const T_gvf     *u,
                   const T_clvf    *fclinv)
#else
template <typename T_solver, typename T_wqf, typename T_gvf>
void
assign_inv_matvec(       T_solver  &solver,
                   const float     *skappa,
                   const float     *stol,
                         int       *iter,
                         T_wqf     *restrict yd,
                   const T_wqf     *restrict  y,
                   const T_gvf     *u)
#endif
{
//
// MAIN MIC Solver with BiCGStab algorithm for 
//
//  D yde = y
//

  startTimer(solver_timer.solv_timer);

#ifdef _DEBUG_
  double t0 = omp_get_wtime();
#endif

  float stol2 = (*stol)*(*stol);
  const int niter = 10000;

  float _ALIG_PRE_ rtmp0 _ALIG_PST_;
  std::complex<float> _ALIG_PRE_ ctmp0,ctmp1 _ALIG_PST_;
  std::complex<float> _ALIG_PRE_  alp, bet, omg, rho0, rho1 _ALIG_PST_;
  std::complex<float> _ALIG_PRE_ nalp, nomg _ALIG_PST_;

  T_wqf *restrict  r;
  T_wqf *restrict rt;
  T_wqf *restrict  p;
  T_wqf *restrict  q;
  T_wqf *restrict  t;
  T_wqf *restrict  v;

  const int VOLUME = VLEN*count(v);

#ifdef _DEBUG_
  fflush(stdout);
#ifdef _CLOVER
  if ( 0 == _nodeid_ ) printf("%%S_BICGSTAB_CLOVER(MIC)\n");
#else
  if ( 0 == _nodeid_ ) printf("%%S_BICGSTAB_WILSON(MIC)\n");
#endif
  if ( 0 == _nodeid_ ) printf("%%  VOLUME = %d\n",VOLUME);
  fflush(stdout);
#endif


   r = wqf_tmp.push();
  rt = wqf_tmp.push();
   p = wqf_tmp.push();
   q = wqf_tmp.push();
   t = wqf_tmp.push();

  if ( 1 == solver.is_prec ) {
   v = wqf_tmp.push();
  }

  assign_attribute( r,y);
  assign_attribute(rt,y);
  assign_attribute( p,y);
  assign_attribute( q,y);
  assign_attribute( t,y);

  if ( 1 == solver.is_prec ) assign_attribute( v,y);


  //
  // set x_0 = 0 (yd = 0)
  //
  clear(yd);
  assign_attribute(yd,y);

  //
  // r  = b ( = y : rhs of equation )
  // rt = r
  // p  = r
  //
  assign( r,y);
  assign(rt,r);
  assign( p,r);

  //
  // rho0 = |re|^2
  //
  rtmp0 = abs2(r);
#ifdef _TIMING_
  solver_timer.solv_flop_count += QABS2*VOLUME;
#endif

  rho0  = rtmp0;

#ifdef _DEBUG_
  float err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%% S_TOL=%24.15e\n",sqrt(stol2));
  if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {

    if ( 1 == solver.is_prec ) {
      //
      // v = Mras p 
      // q = A v 
      //
#ifdef _CLOVER
      solver.precon(skappa,v,p,u,fclinv);
      solver.matvec(skappa,q,v,u,fclinv);
#else
      solver.precon(skappa,v,p,u);
      solver.matvec(skappa,q,v,u);
#endif
    } else {
      //
      // q = A p 
      //
#ifdef _CLOVER
      solver.matvec(skappa,q,p,u,fclinv);
#else
      solver.matvec(skappa,q,p,u);
#endif
    }


    *iter = *iter + 1;

    //
    // alp = rho0/<rt|q>
    //
    ctmp0 = prod(rt,q);
#ifdef _TIMING_
    solver_timer.solv_flop_count += QPROD*VOLUME;
#endif
      alp = rho0/ctmp0;
     nalp = -alp;

    startTimer(solver_timer.update_timer);
    if ( 1 == solver.is_prec ) {
      //
      // x = x + alp v
      // r = r - alp q
      //
      accum_add_mult(yd, &alp,v);
      accum_add_mult( r,&nalp,q);
    } else {
      //
      // x = x + alp p
      // r = r - alp q
      //
      accum_add_mult(yd, &alp,p);
      accum_add_mult( r,&nalp,q);
    }
    stopTimer(solver_timer.update_timer);

    rtmp0 = abs2(r);

#ifdef _TIMING_
    solver_timer.update_flop_count += (2*QACCUM_ADD_CMULT)*VOLUME;
    solver_timer.solv_flop_count += (2*QACCUM_ADD_CMULT+QABS2)*VOLUME;
#endif

#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    if (rtmp0 < stol2) break;

    if ( 1 == solver.is_prec ) {
      //
      // v = Mras r
      // t = A v
      //
#ifdef _CLOVER
      solver.precon(skappa,v,r,u,fclinv);
      solver.matvec(skappa,t,v,u,fclinv);
#else
      solver.precon(skappa,v,r,u);
      solver.matvec(skappa,t,v,u);
#endif
    } else {
      //
      // t = A r
      //
#ifdef _CLOVER
      solver.matvec(skappa,t,r,u,fclinv);
#else
      solver.matvec(skappa,t,r,u);
#endif
    }

    *iter = *iter + 1;

    //
    // omg = <t|r>/<t|t>
    //

    rtmp0 = abs2(t);
    ctmp0 = prod(t,r);

#ifdef _TIMING_
    solver_timer.solv_flop_count += (QABS2+QPROD)*VOLUME;
#endif

      omg = ctmp0/rtmp0;
     nomg = -omg;

    startTimer(solver_timer.update_timer);
    if ( 1 == solver.is_prec ) {
      //
      // x = x + omg v
      // r = r - omg t
      //
      accum_add_mult(yd, &omg,v);
      accum_add_mult( r,&nomg,t);
    } else {
      //
      // x = x + omg r
      // r = r - omg t
      //
      accum_add_mult(yd, &omg,r);
      accum_add_mult( r,&nomg,t);
    }
    stopTimer(solver_timer.update_timer);

    rtmp0 = abs2(r);

#ifdef _TIMING_
    solver_timer.update_flop_count += (2*QACCUM_ADD_CMULT)*VOLUME;
    solver_timer.solv_flop_count += (2*QACCUM_ADD_CMULT+QABS2)*VOLUME;
#endif

#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    if (rtmp0 < stol2) break;

    //
    // rho1 = <rt|r>
    //
    rho1 = prod(rt,r);
#ifdef _TIMING_
    solver_timer.solv_flop_count += (QPROD)*VOLUME;
#endif

    //
    // bet = (alp/omg)*(rho1/rho0)
    // rho0 = rho1;
    //
    ctmp0 = alp/omg;
    ctmp1 = rho1/rho0;
      bet = ctmp0*ctmp1;
     rho0 = rho1;

    startTimer(solver_timer.update_timer);
    //
    // p = r + bet ( p - omg q)
    //
    accum_add_mult(p,&nomg,q);
    accum_mult_add(p, &bet,r);
    stopTimer(solver_timer.update_timer);

#ifdef _TIMING_
    solver_timer.update_flop_count += (QACCUM_ADD_CMULT+QACCUM_CMULT_ADD)*VOLUME;
    solver_timer.solv_flop_count += (QACCUM_ADD_CMULT+QACCUM_CMULT_ADD)*VOLUME;
#endif

  } // end while

  //======================================
  // check iteration counts (max iter)
  // if iter exceeds max iteration
  // flip the sign.
  //======================================
  if (*iter >= niter) { *iter = -(*iter); }

#ifdef _DEBUG_
  //======================================
  // Check residual
  //
  // qe = Dee yde
  //
#ifdef _CLOVER
  solver.matvec(skappa,q,yd,u,fclinv);
#else
  solver.matvec(skappa,q,yd,u);
#endif
  //
  // re = ye - qe
  //
  assign_sub(r,y,q);

  rtmp0 = abs2(r);

#ifdef _TIMING_
  solver_timer.solv_flop_count += (QSUB+QABS2)*VOLUME;
#endif

  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%%%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
  //======================================
#endif

  wqf_tmp.pop(5);
  if ( 1 == solver.is_prec ) wqf_tmp.pop(1);

#ifdef _DEBUG_
  double t1 = omp_get_wtime();
  if ( 0 == _nodeid_ ) printf("%%ETIME_=%24.15E\n",(t1-t0));
  fflush(stdout);
#endif

  stopTimer(solver_timer.solv_timer);

}

#endif
