///////////////////////////////////////////////////////////////////
// MIC native solver interfaces.
// This calles the MIC native solver code in the offload region
// The interfaces can be called from Fortran
///////////////////////////////////////////////////////////////////

#include <immintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>
#include "cconfig.h"
#include "cmic_config.h"

#include "mic_nodeinfo.h"
#include "mic_param_size.h"

//////////////////////////////////////////////////
// node infor in the host side
// The value must be set in the fortran module.
//////////////////////////////////////////////////
extern "C" mic_nodeinfo m_host_nodeinfo;

// interface routines for offloaded solver
#include <mpi.h>
#include "signal.h"

///////////////////////////////////////////////
// host offload solver data structure
//   "host_ras_solver" is defined
///////////////////////////////////////////////
#include "mic_offload_solver_class.h"

#pragma offload_attribute(push, target(mic))
#include "scif_interface.h"
#pragma offload_attribute(pop)

///////////////////////////////////////////////
// reverse offload interface 
///////////////////////////////////////////////
#include "mic_reverse_offload.h"

/////////////////////////////////////////////////////////////
// MIC native code prototypes
/////////////////////////////////////////////////////////////
extern "C" __declspec(target(mic)) void 
new_mic_ras_solver(const mic_nodeinfo    *ni,
                   const mic_gvf_soa     *u,
                   const mic_ras_gvf_soa *ru,
                   const mic_clvf        *fclinv,
                   const mic_ras_clvf    *rfclinv);

extern "C" __declspec(target(mic)) void
delete_mic_ras_solver();

extern "C" __declspec(target(mic)) void
assign_inv_mult_eoprec_wd_bicgstab_mic(const float  *skappa,
                                       const float  *stol,
                                       int          *iter,
                                             mic_wqf_eo   *restrict yde,
                                       const mic_wqf_eo   *restrict  ye);
/////////////////////////////////////////////////////////////
// END MIC native code prototypes
/////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////
// OFFLOAD INTERFACES
/////////////////////////////////////////////////////////////
__declspec(target(mic)) static mic_gvf_soa     *u1;
__declspec(target(mic)) static mic_ras_gvf_soa *u_ras1;
__declspec(target(mic)) static mic_clvf        *fclinv1;
__declspec(target(mic)) static mic_ras_clvf    *fclinv_ras1;

extern "C" void
new_offl_ras_solver(const mic_gvf_soa *u,
                    const mic_ras_gvf_soa *u_ras,
                    const mic_clvf *fclinv,
                    const mic_ras_clvf *fclinv_ras)
{
    mic_nodeinfo ni = m_host_nodeinfo;

    if (u1 == 0) {
    // Allocate the target struct and its pointers
    #pragma offload target(mic:mic_targetid) \
      nocopy(u1         [0:1] : alloc_if(1) free_if(0) preallocated targetptr) \
      nocopy(u_ras1     [0:1] : alloc_if(1) free_if(0) preallocated targetptr) \
      nocopy(fclinv1    [0:1] : alloc_if(1) free_if(0) preallocated targetptr) \
      nocopy(fclinv_ras1[0:1] : alloc_if(1) free_if(0) preallocated targetptr)
    {
      u1 = (mic_gvf_soa *)_mm_malloc(sizeof(mic_gvf_soa), 64);
      u_ras1 = (mic_ras_gvf_soa *)_mm_malloc(sizeof(mic_ras_gvf_soa), 64);
      fclinv1  = (mic_clvf *)_mm_malloc(sizeof(mic_clvf), 64);
      fclinv_ras1  = (mic_ras_clvf *)_mm_malloc(sizeof(mic_ras_clvf), 64);
    }
    }

    // Transfer the data to the card
    #pragma offload target(mic:mic_targetid) \
        in(u          [0:1] : into(u1         [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(u_ras      [0:1] : into(u_ras1     [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(fclinv     [0:1] : into(fclinv1    [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(fclinv_ras [0:1] : into(fclinv_ras1[0:1]) alloc_if(0) free_if(0) targetptr) \
        in(ni)
    {
        new_mic_ras_solver(&ni,u1,u_ras1,fclinv1,fclinv_ras1);
    }
}


extern "C" void
delete_offl_ras_solver()
{
    void *u1, *u_ras1, *fclinv1, *fclinv_ras1;
    mic_gvf_soa     *u;
    mic_ras_gvf_soa *u_ras;
    mic_clvf        *fclinv;
    mic_ras_clvf    *fclinv_ras;

#if 1
    #pragma offload target(mic:mic_targetid)
    { delete_mic_ras_solver(); }
#else
    // Get the device pointers
    #pragma offload target(mic:mic_targetid) \
        in(mic_ras_ptr[0:0] : alloc_if(0) free_if(0) targetptr)
    {
        u1          = (void *)mic_ras_ptr->u;  // restore gauge field device pointers from the mic_ras_ptr
        u_ras1      = (void *)mic_ras_ptr->u_ras;
        fclinv1     = (void *)mic_ras_ptr->fclinv;
        fclinv_ras1 = (void *)mic_ras_ptr->fclinv_ras;
        delete_mic_ras_solver();
    }

    u          = (mic_gvf_soa *)u1;
    u_ras      = (mic_ras_gvf_soa *)u_ras1;
    fclinv     = (mic_clvf *)fclinv1;
    fclinv_ras = (mic_ras_clvf *)fclinv_ras1;

    // free the device memory
    #pragma offload_transfer target(mic:mic_targetid) \
        in(mic_ras_ptr[0:0] : alloc_if(0) free_if(1) targetptr) \
        in(u          [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(u_ras      [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(fclinv     [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(fclinv_ras [0:0] : alloc_if(0) free_if(1) targetptr)
#endif
}

#include "commands.h"

extern "C" void
assign_inv_mult_offl_ras_solver( const float      *kappa,
                                 const float      *stol,
                                       int        *iter,
                                 const mic_wqf_eo *swe,
                                       mic_wqf_eo *sve)
{

    int offload_signal = 0xF;
    const float skappa1 = *kappa;
    const float stol1 = *stol;
    int iter1 = *iter;

    ////////////////////////////////////////////////////////////////////////
    // Asynchronous offloading of native code
    //   proxy server is running parallel to the native code.
    ////////////////////////////////////////////////////////////////////////
#if 1
    static mic_wqf_eo *swe1, *sve1;
    if (swe1 == 0)
    {
    #pragma offload target(mic:mic_targetid) \
      nocopy(swe1[0:1] : alloc_if(1) free_if(0) preallocated targetptr) \
      nocopy(sve1[0:1] : alloc_if(1) free_if(0) preallocated targetptr)
    {
      swe1 = (mic_wqf_eo *)_mm_malloc(sizeof(mic_wqf_eo), 64);
      sve1 = (mic_wqf_eo *)_mm_malloc(sizeof(mic_wqf_eo), 64);
    }
    }
      
    #pragma offload target(mic:mic_targetid) \
        nocopy(scif_mic) \
        in( swe [0:1] : into(swe1[0:1]) alloc_if(0) free_if(0) targetptr) \
        out(sve1[0:1] : into(sve [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(skappa1,stol1) \
        inout(iter1) \
        signal(offload_signal)
    {

        assign_inv_mult_eoprec_wd_bicgstab_mic(&skappa1,&stol1,&iter1,sve1,swe1);

#ifndef _singlePU
        //////////////////////////////////////
        // send Stop-message to the proxy 
        //////////////////////////////////////
        Command cmd;
        cmd.cmd = Command::CMD_Stop;
        cmd.cmdlen = sizeof(Command);
        scif_mic->enqueue(&cmd, sizeof(cmd), true);
        scif_mic->qempty(true);
#endif
#else
    #pragma offload target(mic:mic_targetid) \
        nocopy(scif_mic) \
        in(  swe[0:1] : alloc_if(1) free_if(1)) \
        out( sve[0:1] : alloc_if(1) free_if(1)) \
        in(skappa1,stol1) \
        inout(iter1) \
        signal(offload_signal)
    {

        assign_inv_mult_eoprec_wd_bicgstab_mic(&skappa1,&stol1,&iter1,sve,swe);

#ifndef _singlePU
        //////////////////////////////////////
        // send Stop-message to the proxy 
        //////////////////////////////////////
        Command cmd;
        cmd.cmd = Command::CMD_Stop;
        cmd.cmdlen = sizeof(Command);
        scif_mic->enqueue(&cmd, sizeof(cmd), true);
        scif_mic->qempty(true);
#endif
#endif
    }

    ////////////////////////////////////////////////
    // process proxy 
    // receive requests from the native code located
    // in the above offload region.
    ////////////////////////////////////////////////
    #ifndef _singlePU
    process_cmds();
    #endif

    ////////////////////////////////////////////////
    // wait for finishing of offloading
    ////////////////////////////////////////////////
    #pragma offload_wait target(mic:mic_targetid) wait(offload_signal)

    *iter = iter1;
}

