#include <immintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>
#include "cconfig.h"
#include "cmic_config.h"
#include "mic_nodeinfo.h"
#include "mic_param_size.h"
#include <mpi.h>
#include "signal.h"

/////////////////////////////////////////////////////////////////////////////////
// interface routines for offloaded solver with reverse-offloading
/////////////////////////////////////////////////////////////////////////////////

#pragma offload_attribute(push, target(mic))
#include "scif_interface.h"
#pragma offload_attribute(pop)


///////////////////////////////////////////
// SCIF interface used in the host side
///////////////////////////////////////////
SCIF_Interface *scif_host;

///////////////////////////////////////////
// SCIF interface used in the MIC side
///////////////////////////////////////////
__declspec(target(mic)) 
SCIF_Interface *scif_mic;

///////////////////////////////////////////
// MIC device id used in the host side 
// to launch the offload code
///////////////////////////////////////////
int mic_targetid;

///////////////////////////////////////////
// Affinity setting used in the MIC code
///////////////////////////////////////////
__declspec(target(mic)) 
char *mic_affinity;

///////////////////////////////////////////
// signal handlers
///////////////////////////////////////////
__declspec(target(mic)) void
segv_handler(int sig)
{
    printf("%d: MIC: segv_handler: %d\n", getpid(), sig);
    fflush(0);
    _exit(1);
}

void
sigint_handler(int sig)
{
    printf("Host: caught sigint\n");
    fflush(0);
    if (scif_host)
        scif_host->close();
    _exit(1);
}

///////////////////////////////////////////////
// Find the MIC device id and affinity
///////////////////////////////////////////////
extern "C" void
parsecards()
{
#ifndef _singlePU
    int rank;
    int nranks;
    MPI_Comm_size(MPI_COMM_WORLD, &nranks);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    int mics[nranks];
    char *affinities[nranks];
    char *p = getenv("CCS_CARDS");
    if (p == 0)
        for (int i = 0; i < nranks; ++i)
            mics[i] = 0;
    else
    {
        char cards[strlen(p)+1];
        strcpy(cards, p);
        int i = 0;
        char *q;
        for (q = strtok(cards, ","); q != 0; q = strtok(0, ","))
        {
            if (i >= nranks)
                break;
            int card = atoi(q);
            mics[i++] = card;
        }
        int lastrank = i;
        i = 0;
        for (int j = lastrank; j < nranks; ++j)
        {
            mics[j] = mics[i];
            ++i;
            if (i == lastrank)
                i = 0;
        }
    }
    mic_targetid = mics[rank];

#ifdef CCS_AFFINITIES
    p = getenv("CCS_AFFINITIES");
    if (p == 0)
        for (int i = 0; i < nranks; ++i)
            affinities[i] = 0;
    else
    {
        p = strdup(p);
        int i = 0;
        char *q;
        for (q = strtok(p, ","); q != 0; q = strtok(0, ","))
        {
            if (i >= nranks)
                break;
            printf("q = %s\n", q);
            affinities[i++] = q;
        }
        int lastrank = i;
        i = 0;
        for (int j = lastrank; j < nranks; ++j)
        {
            affinities[j] = affinities[i];
            ++i;
            if (i == lastrank)
                i = 0;
        }
    }
    mic_targetid = mics[rank];
    mic_affinity = affinities[rank];
    printf("%d: MIC TARGET %d affinity %s\n", rank, mic_targetid, mic_affinity);
#endif
#endif
}

//////////////////////////////////////////////////////////////////
// start up the SCIF interfaces on both the MIC and host 
//////////////////////////////////////////////////////////////////
extern "C" void
start_scif()
{
    int offload_signal = 0xF;
    int mic_node;
    uint16_t mic_port;
    parsecards();

#if 1
    if (mic_affinity == 0)
        mic_affinity = "";
    #pragma offload target(mic:mic_targetid) nocopy(scif_mic) \
        in(mic_affinity:length(strlen(mic_affinity)+1)) \
        out(mic_node) out(mic_port)
    {
        signal(SIGSEGV, segv_handler);
        char aff[1024];
        if (*mic_affinity != 0)
        {
            snprintf(aff, sizeof(aff), "KMP_PLACE_THREADS=%s", mic_affinity);
            kmp_set_defaults(aff);
        }
        printf("MIC creating server port %s\n", aff);
        scif_mic = new SCIF_Interface();
        mic_port = scif_mic->create_receive_port(mic_node);
        printf("MIC created server port\n");
    }
#else
    #pragma offload target(mic:mic_targetid) nocopy(scif_mic) \
        out(mic_node) out(mic_port)
    {
        signal(SIGSEGV, segv_handler);
        scif_mic = new SCIF_Interface();
        mic_port = scif_mic->create_receive_port(mic_node);
        printf("MIC created server port\n");
    }
#endif
    printf("Host: mic_node = %d mic_port = %d\n", mic_node, mic_port);
    // Start MIC server
    #pragma offload target(mic:mic_targetid) nocopy(scif_mic) \
        signal(offload_signal)
    {
        printf("MIC waiting for connection...\n"); fflush(0);
        scif_mic->server();
        printf("MIC connect successful\n");
    }
    scif_host = new SCIF_Interface();
    scif_host->create_send_port(mic_node, mic_port);
    printf("CPU connected to MIC.\n");
    signal(SIGINT, sigint_handler);
    #pragma offload_wait target(mic:mic_targetid) wait(offload_signal)
}

#include "commands.h"

//////////////////////////////////////////////////////////////////
// proxy that connects the communication request from the MIC 
// region to the MPI calls in the host
//////////////////////////////////////////////////////////////////
extern "C" void
process_cmds()
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    Command cmd1[MaxCmd];
    assert((sizeof(Cmddata) % sizeof(Command)) == 0);
    assert(sizeof(cmd1) >= sizeof(Command) + sizeof(Cmddata));
    memset(cmd1, 0, sizeof(cmd1));
    int nreqs = 0;
    MPI_Request reqs[MaxReq];
    off_t dirtybuf[MaxReq];
    int dirtylen[MaxReq];
    for (;;)
    {
        uint64_t msgstart;
        size_t len;
        int data;
        int retry = 0;
        for (;;)
        {
            scif_host->dequeue(data, msgstart, len, false);
            if (len != 0)
                break;
            // Check request status
            ++retry;
            for (int i = 0; i < nreqs; ++i)
            {
                int flag = 0;
                MPI_Test(&reqs[i], &flag, MPI_STATUS_IGNORE);
                if (flag && dirtybuf[i])
                {
                    scif_host->writebuf(dirtybuf[i], dirtylen[i]);
                    dirtybuf[i] = 0;
                    dirtylen[i] = 0;
                    retry = 0;
                }
            }
            if ((retry % 1000) == 0)
                scif_host->checkep();
        }

        assert(len <= sizeof(cmd1));
        scif_host->qcopy(cmd1, msgstart, len);
        //printf("%d: Command: %d len %lu\n", rank, cmd1->cmd, len);
        if (cmd1->cmd == Command::CMD_Stop)
        {
            break;
        }
        Cmddata cmddata;
        memcpy(&cmddata, cmd1->data, sizeof(cmddata));
        void *buff1, *buff2;
        MPI_Status status;
        switch (cmd1->cmd)
        {
        case CMD_Isend:
            scif_host->readbuf(cmddata.sbuff, cmddata.buflen);
            scif_host->fence();
            buff1 = scif_host->getbuf(cmddata.sbuff);
            assert(nreqs < MaxReq);
            MPI_Isend(buff1, cmddata.buflen, MPI_BYTE, cmddata.snode, 0,
                MPI_COMM_WORLD, &reqs[nreqs]);
            dirtybuf[nreqs] = 0;
            dirtylen[nreqs] = 0;
            ++nreqs;
            break;
        case CMD_Irecv:
            buff1 = scif_host->getbuf(cmddata.rbuff);
            assert(nreqs < MaxReq);
            MPI_Irecv(buff1, cmddata.buflen, MPI_BYTE, cmddata.rnode, 0,
                MPI_COMM_WORLD, &reqs[nreqs]);
            dirtybuf[nreqs] = cmddata.rbuff;
            dirtylen[nreqs] = cmddata.buflen;
            ++nreqs;
            break;
        case CMD_Waitall:
            MPI_Waitall(nreqs, reqs, MPI_STATUSES_IGNORE);
            for (int i = 0; i < nreqs; ++i)
                if (dirtybuf[i])
                    scif_host->writebuf(dirtybuf[i], dirtylen[i]);
            nreqs = 0;
            scif_host->fence();
            break;
        case CMD_Fence:
            scif_host->fence();
            break;
        case CMD_Sendrecv:
            scif_host->readbuf(cmddata.sbuff, cmddata.buflen);
            scif_host->fence();
            buff1 = scif_host->getbuf(cmddata.sbuff);
            buff2 = scif_host->getbuf(cmddata.rbuff);
            MPI_Sendrecv(buff1, cmddata.buflen, MPI_BYTE, cmddata.snode, 0,
                         buff2, cmddata.buflen, MPI_BYTE, cmddata.rnode, 0,
                         MPI_COMM_WORLD, &status);
            scif_host->writebuf(cmddata.rbuff, cmddata.buflen);
            scif_host->fence();
            break;
        case CMD_Allreduce:
            scif_host->readbuf(cmddata.sbuff, cmddata.buflen);
            scif_host->fence();
            buff1 = scif_host->getbuf(cmddata.sbuff);
            buff2 = scif_host->getbuf(cmddata.rbuff);
            MPI_Allreduce(buff1, buff2, cmddata.buflen/sizeof(float),
                MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
            scif_host->writebuf(cmddata.rbuff, cmddata.buflen);
            scif_host->fence();
            break;
        case CMD_Nop:
            break;
        default:
            assert(0);
        }
    }
}
