#ifndef _STAT_H
#define _STAT_H
#include <stdint.h>
#include <omp.h>
class Stat
{
    const char *name;
    __declspec(align(64)) uint64_t ctrs[8][256];
    // cumulative counters
    uint64_t count;     // number of invocations
    uint64_t tregion;   // total time in parallel region (from thread 0)
    uint64_t tfork;     // max fork time
    uint64_t tjoin;     // max imbalance + join overhead
    uint64_t tcycles;   // total cycles inside parallel region
    uint64_t pmc0, pmc1;// pmu
    // add memory counters here
    // temp variables
    uint64_t tstart;    // tsc at start of parallel region
    uint64_t tend;      // tsc at end of parallel region
    // map for ctrs values
    // 0 pmc0 start
    // 1 pmc0 end
    // 2 pmc1 start
    // 3 pmc1 end
    // 4 tsc start
    // 5 tsc end
public:
    void init(const char *regname)
    {
        name = regname;
        clear();
    }
    void clear()
    {
        count = 0;
        tregion = 0;
        tfork = 0;
        tjoin = 0;
        pmc0 = 0;
        pmc1 = 0;
        tcycles = 0;
    }
    void print()
    {
        printf(
            "Reg %s: region %lu fork %lu join %lu "
            "cycles %lu pmc0 %lu pmc1 %lu count %d\n",
            name, tregion, tfork, tjoin, tcycles, pmc0, pmc1, count);
    }
    void start()
    {
        ++count;
        tstart = _rdtsc();
    }
    void enter(int t)
    {
        ctrs[0][t] = _rdpmc(0);
        ctrs[2][t] = _rdpmc(1);
        ctrs[4][t] = _rdtsc();
    }
    void exit(int t)
    {
        ctrs[1][t] = _rdpmc(0);
        ctrs[3][t] = _rdpmc(1);
        ctrs[5][t] = _rdtsc();
    }
    void accum(int nthreads)
    {
        tend = _rdtsc();
        for (int i = 0; i < nthreads; ++i) {
            uint64_t p0 = ctrs[1][i] - ctrs[0][i];
            uint64_t p1 = ctrs[3][i] - ctrs[2][i];
            uint64_t c  = ctrs[5][i] - ctrs[4][i];
            uint64_t fork = ctrs[4][i] - tstart;
            uint64_t join = tend - ctrs[5][i];
            pmc0 += p0;
            pmc1 += p1;
            tcycles += c;
            if (tfork < fork) tfork = fork;
            if (tjoin < join) tjoin = join;
        }
        uint64_t region = tend - tstart;
        tregion += region;
    }
};
#endif
